import { Response } from 'express';
import * as iconv from 'iconv-lite';
import * as moment from 'moment';
import { Readable } from 'stream';
import * as stopWordsFr from './stop-words-fr.json';

export type ValuesOf<T extends any[]> = T[number];
export type GroupKey<T, K extends keyof T = any> = {
  key: Pick<T, K>;
  items: T[];
};

export class Utils {
  public static alphabeticalPattern: RegExp = /[a-zA-Zéèàùâûôêîç '-]*/;
  public static readonly PasswordPattern = new RegExp('^(?=.*[0-9])(?=.*[a-z])(?=.*[A-Z])(?=.*[*.!@$%^&(){}[]:;<>,.?/~_+-=|]*).{8,}$');

  public static extractCommonNameFromDN(dn?: string): string | undefined {
    if (!dn || dn.length < 3) {
      return undefined;
    }
    return dn.substring(3, dn.indexOf(','));
  }

  public static escapeSqlParameter<T>(value: T) {
    if (null == value) return 'NULL';
    if (Array.isArray(value)) {
      var vals = value.map(Utils.escapeSqlParameter);
      return '(' + vals.join(', ') + ')';
    }
    let val: string = '' + value;
    const backslash: number = ~val.indexOf('\\');
    const prefix = backslash ? 'E' : '';
    val = val.replace(/'/g, "''");
    val = val.replace(/\\/g, '\\\\');
    return prefix + "'" + val + "'";
  }

  public static addFileDownloadToResponse(response: Response, buffer: Buffer, filename: string): void {
    response.header('Content-disposition', 'attachment; filename=' + filename);

    const stream: Readable = new Readable();
    stream.push(buffer);
    stream.push(null);
    stream.pipe(response);
  }

  public static encode(str: string): Buffer {
    return iconv.encode(str, 'WINDOWS-1252');
  }

  public static encodeISO_8859_1(str: string): Buffer {
    return iconv.encode(str, 'ISO-8859-1');
  }

  public static isNumber(value: any): value is number {
    return typeof value === 'number';
  }

  public static isString(value: any): value is string {
    return typeof value === 'string';
  }

  public static isNotEmpty(value: any): boolean {
    return value !== null && value !== undefined && value !== '';
  }

  public static isNullOrEmpty(value: any): boolean {
    return value == null || value === '';
  }

  public static flatDeep<T = any>(arr: T[]): T[] {
    return arr.reduce((acc: T[], val: T | T[]) => acc.concat(Array.isArray(val) ? Utils.flatDeep(val) : val), []);
  }

  public static normalize<TEntity>(target: TEntity, source: TEntity): void {
    if (target && source) {
      Object.keys(source).forEach((key: string) => {
        target[key] = source[key] !== undefined ? source[key] : null;
      });
    }
  }

  public static normalizeStupid<TEntity>(target: TEntity, source: TEntity): void {
    if (target && source) {
      Object.keys(source).forEach((key: string) => {
        target[key] = source[key];
      });
    }
  }

  public static normalizeString(str: string): string | undefined {
    if (str) {
      return str.normalize('NFD').replace(/[\u0300-\u036f]/g, '');
    }
  }

  public static normalizeFilename(str: string): string | undefined {
    if (str) {
      return str.replace(/ /g, '_');
    }
  }

  public static range(start: number, end: number): number[] {
    return Array(end - start + 1)
      .fill(start)
      .map((value: number, index: number) => value + index);
  }

  public static groupByKey<T, A extends Array<keyof T>, K extends ValuesOf<A>>(items: T[] = [], ...keys: A): GroupKey<T, K>[] {
    return items.reduce((groups: GroupKey<T, K>[], item: T) => {
      const key: Pick<T, K> = Utils.pick(item, keys);
      const index: number = groups.findIndex((value: GroupKey<T, K>) => JSON.stringify(value.key) === JSON.stringify(key));
      if (index === -1) {
        groups.push({
          key,
          items: [item],
        });
      } else {
        groups[index].items.push(item);
      }

      return groups;
    }, []);
  }

  public static groupBy<T, Fn extends (item: T) => string | number>(items: T[], fn: Fn): { [K in keyof any]: T[] } {
    return items.reduce((groups: { [K in keyof any]: T[] }, item: T) => {
      const key: string | number = fn(item);
      if (!groups[key]) {
        groups[key] = [];
      }
      groups[key].push(item);

      return groups;
    }, {});
  }

  public static pick<T, K extends keyof T>(source: T, keys: K[]): Pick<T, K> {
    return keys.reduce((acc: any, key: K) => {
      acc[key] = source[key];
      return acc;
    }, {});
  }

  public static sleep(ms: number): Promise<void> {
    return new Promise((resolve) => {
      setTimeout(resolve, ms);
    });
  }

  public static parseBool(value: any): boolean {
    if (value === true || value === 'true') {
      return true;
    }
    if (value === false || value === 'false') {
      return false;
    }
  }

  public static normalizeSearch(search: string): string[] {
    search = search.replace(`'`, ` `);
    let words = search.split(` `).filter((word) => {
      return !stopWordsFr.includes(word.toLocaleLowerCase());
    });
    words.forEach((word, i) => {
      words[i] = Utils.splitAndJoin(word);
      if (words[i] === '%%') {
        delete words[i];
      }
    });
    return words;
  }

  public static splitAndJoin(value: string): string {
    const search: string = value.split('*').join('%');
    if (search.search('%') === -1) {
      return '%' + search + '%';
    }
    return search;
  }

  public static getTimeFromMins(mins: number): string {
    const h = (mins / 60) | 0;
    const m = mins % 60 | 0;
    return moment.utc().hours(h).minutes(m).format('H[h]mm');
  }

  /**
   * Transform a phone number from '0123456789' to '01 23 45 67 89'.
   */
  public static formatPhone(phone: string, separator: string = ' '): string {
    // Transform '0123456789' to '01 23 45 67 89'
    // Cf. https://stackoverflow.com/a/8358141
    const cleaned: string = phone?.trim();
    const match: RegExpMatchArray = cleaned?.match(/^(\d{2})(\d{2})(\d{2})(\d{2})(\d{2})$/);
    if (match) {
      return match.slice(1).join(separator); // Slicing because match[0] is the whole initial phone number
    }
    return phone; // Format did not succeed, so returning initial specified phone
  }

  /**
   * Format bytes as human-readable text.
   * E.g. 1200000 return '1.2 MB'.
   *
   * Cf. https://stackoverflow.com/a/14919494
   *
   * @param bytes Number of bytes
   * @param si True to use metric (SI) units, aka powers of 1000. False to use binary (IEC), aka powers of 1024
   * @param dp Number of decimal places to display
   *
   * @return Formatted string
   */
  public static humanFileSize(filesize: number, si: boolean = false, dp: number = 1): string {
    const thresh: number = si ? 1000 : 1024;

    if (filesize === undefined || filesize === null) {
      return <any>filesize;
    }

    if (Math.abs(filesize) < thresh) {
      return filesize + ' octet' + (filesize ? 's' : '');
    }

    const units = si ? ['ko', 'Mo', 'Go', 'To', 'Po', 'Eo', 'Zo', 'Yo'] : ['Kio', 'Mio', 'Gio', 'Tio', 'Pio', 'Eio', 'Zio', 'Yio'];
    let u: number = -1;
    const r: number = 10 ** dp;
    do {
      filesize /= thresh;
      ++u;
    } while (Math.round(Math.abs(filesize) * r) / r >= thresh && u < units.length - 1);

    return Number(filesize.toFixed(dp)) + ' ' + units[u]; // Using Number(...) to remove insignificant trailing zeros, Cf. https://stackoverflow.com/a/19623253
  }
}
